#include "container.h"

using namespace std;

namespace Msp {
namespace GLtk {

hierarchy_error::hierarchy_error(const string &w):
	logic_error(w)
{ }


Container::Container():
	click_focus(0),
	click_button(0),
	pointer_focus(0),
	pointer_grabbed(false),
	input_focus(0)
{ }

Container::~Container()
{
	while(!children.empty())
		delete children.front()->widget;
}

void Container::add(Widget &wdg)
{
	wdg.set_parent(this);
	children.push_back(create_child(&wdg));
	on_child_added(wdg);
}

void Container::remove(Widget &wdg)
{
	for(list<Child *>::iterator i=children.begin(); i!=children.end(); ++i)
		if((*i)->widget==&wdg)
		{
			wdg.set_parent(0);
			delete *i;
			children.erase(i);
			on_child_removed(wdg);
			return;
		}

	throw hierarchy_error("widget not in container");
}

Container::Child *Container::create_child(Widget *wdg)
{
	return new Child(*this, wdg);
}

list<Widget *> Container::get_children() const
{
	list<Widget *> result;
	for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
		result.push_back((*i)->widget);
	return result;
}

Widget *Container::get_child_at(int x, int y)
{
	for(list<Child *>::iterator i=children.end(); i!=children.begin();)
		if((*--i)->widget->is_visible() && (*i)->widget->get_geometry().is_inside(x, y))
			return (*i)->widget;

	return 0;
}

Widget *Container::get_descendant_at(int x, int y)
{
	Widget *wdg = get_child_at(x, y);
	if(Container *cont = dynamic_cast<Container *>(wdg))
	{
		const Geometry &cgeom = wdg->get_geometry();
		Widget *wdg2 = cont->get_descendant_at(x-cgeom.x, y-cgeom.y);
		if(wdg2)
			return wdg2;
	}
	return wdg;
}

void Container::raise(Widget &wdg)
{
	for(list<Child *>::iterator i=children.begin(); i!=children.end(); ++i)
		if((*i)->widget==&wdg)
		{
			children.splice(children.end(), children, i);
			return;
		}

	throw hierarchy_error("widget not in container");
}

void Container::set_pointer_focus(Widget *wdg)
{
	if(wdg!=pointer_focus)
	{
		if(pointer_focus)
			pointer_focus->pointer_leave();

		pointer_focus = wdg;

		if(pointer_focus)
			pointer_focus->pointer_enter();
	}
}

void Container::set_input_focus(Widget *wdg)
{
	if(wdg!=input_focus)
	{
		if(input_focus)
			input_focus->focus_out();

		input_focus = wdg;

		if(input_focus)
		{
			raise(*wdg);
			input_focus->focus_in();
		}
	}
}

Widget *Container::get_final_input_focus() const
{
	if(Container *container = dynamic_cast<Container *>(input_focus))
		if(Widget *focus = container->get_final_input_focus())
			return focus;

	return input_focus;
}

void Container::button_press(int x, int y, unsigned btn)
{
	if(Widget *child = get_pointer_target(x, y))
	{
		if(!click_focus)
		{
			set_pointer_focus(child);
			if(child->is_focusable())
				set_input_focus(child);

			click_focus = child;
			click_button = btn;
		}

		const Geometry &cgeom = child->get_geometry();
		child->button_press(x-cgeom.x, y-cgeom.y, btn);
	}
}

void Container::button_release(int x, int y, unsigned btn)
{
	if(Widget *child = get_pointer_target(x, y))
	{
		if(child==click_focus && btn==click_button)
		{
			click_focus = 0;
			if(!pointer_focus)
				set_pointer_focus(get_child_at(x, y));
		}

		const Geometry &cgeom = child->get_geometry();
		child->button_release(x-cgeom.x, y-cgeom.y, btn);
	}
}

void Container::pointer_motion(int x, int y)
{
	Widget *child = get_pointer_target(x, y);
	if(!pointer_grabbed)
		set_pointer_focus((!click_focus || child->get_geometry().is_inside(x, y)) ? child : 0);

	if(child)
	{
		const Geometry &cgeom = child->get_geometry();
		child->pointer_motion(x-cgeom.x, y-cgeom.y);
	}
}

Widget *Container::get_pointer_target(int x, int y)
{
	if(pointer_grabbed)
		return pointer_focus;
	else if(click_focus)
		return click_focus;
	else
		return get_child_at(x, y);
}

void Container::pointer_leave()
{
	Widget::pointer_leave();
	set_pointer_focus(0);
}

void Container::key_press(unsigned key, unsigned mod)
{
	if(input_focus)
		input_focus->key_press(key, mod);
}

void Container::key_release(unsigned key, unsigned mod)
{
	if(input_focus)
		input_focus->key_release(key, mod);
}

void Container::character(wchar_t ch)
{
	if(input_focus)
		input_focus->character(ch);
}

void Container::focus_out()
{
	set_input_focus(0);
	Widget::focus_out();
}

void Container::on_reparent()
{
	for(list<Child *>::iterator i=children.begin(); i!=children.end(); ++i)
	{
		if(Container *c = dynamic_cast<Container *>((*i)->widget))
			c->on_reparent();
		(*i)->widget->update_style();
	}
}


Container::Child::Child(Container &c, Widget *w):
	container(c),
	widget(w)
{
	widget->signal_visibility_changed.connect(sigc::mem_fun(this, &Child::visibility_changed));
	widget->signal_request_focus.connect(sigc::mem_fun(this, &Child::request_focus));
	widget->signal_grab_pointer.connect(sigc::mem_fun(this, &Child::grab_pointer));
	widget->signal_ungrab_pointer.connect(sigc::mem_fun(this, &Child::ungrab_pointer));
}

Container::Child::~Child()
{
	visibility_changed(false);
}

void Container::Child::visibility_changed(bool v)
{
	if(!v)
	{
		if(widget==container.click_focus)
			container.click_focus = 0;
		if(widget==container.pointer_focus)
			container.set_pointer_focus(0);
		if(widget==container.input_focus)
			container.set_input_focus(0);
	}
}

void Container::Child::request_focus()
{
	container.set_input_focus(widget);
	if(container.parent && container.visible)
		container.set_focus();
}

void Container::Child::grab_pointer()
{
	if(!container.pointer_grabbed)
	{
		container.set_pointer_focus(widget);
		container.pointer_grabbed = true;
		container.signal_grab_pointer.emit();
	}
}

void Container::Child::ungrab_pointer()
{
	if(container.pointer_grabbed && container.pointer_focus==widget)
	{
		// XXX Should set to the widget under pointer
		container.set_pointer_focus(0);
		container.pointer_grabbed = false;
		container.signal_ungrab_pointer.emit();
	}
}

} // namespace GLtk
} // namespace Msp
