#include <msp/gl/matrix.h>
#include <msp/strings/format.h>
#include "container.h"
#include "resources.h"
#include "root.h"
#include "widget.h"

using namespace std;

namespace Msp {
namespace GLtk {

Widget::Widget():
	style(0),
	state(NORMAL),
	visible(true),
	focusable(true),
	parent(0)
{ }

Widget::~Widget()
{
	if(parent)
	{
		Container *p = parent;
		parent = 0;
		p->remove(*this);
	}
}

void Widget::set_position(int x, int y)
{
	set_geometry(Geometry(x, y, geom.w, geom.h));
}

void Widget::set_size(unsigned w, unsigned h)
{
	set_geometry(Geometry(geom.x, geom.y, w, h));
}

void Widget::autosize()
{
	if(!style)
		return;

	geom.w = 0;
	geom.h = 0;
	const Style::PartSeq &parts = style->get_parts();
	for(Style::PartSeq::const_iterator i=parts.begin(); i!=parts.end(); ++i)
		if(i->get_name().empty())
		{
			const Geometry &pgeom = i->get_geometry();
			const Sides &pmargin = i->get_margin();
			geom.w = max(geom.w, pgeom.w+pmargin.left+pmargin.right);
			geom.h = max(geom.h, pgeom.h+pmargin.top+pmargin.bottom);
		}
}

void Widget::set_geometry(const Geometry &g)
{
	geom = g;
	on_geometry_change();
	rebuild();
}

void Widget::set_parent(Container *p)
{
	if(parent && p)
		throw hierarchy_error("widget already parented");
	else if(p==parent)
		return;

	try
	{
		parent = p;

		on_reparent();
		update_style();
	}
	catch(...)
	{
		// The container has not yet added the widget as its child
		parent = 0;
		throw;
	}
}

void Widget::set_style(const string &s)
{
	style_name = s;
	update_style();
}

void Widget::update_style()
{
	Widget *top;
	for(top=this; top->parent; top=top->parent) ;
	Root *root = dynamic_cast<Root *>(top);
	if(!root)
		style = 0;
	else
	{
		string sname = get_class();
		if(!style_name.empty())
		{
			sname += '-';
			sname += style_name;
		}

		style = &root->get_resources().get<Style>(sname);
	}

	on_style_change();
	signal_autosize_changed.emit();
	rebuild();
}

void Widget::set_tooltip(const string &t)
{
	tooltip = t;
}

void Widget::set_visible(bool v)
{
	if(v==visible)
		return;

	visible = v;

	signal_visibility_changed.emit(visible);
}

void Widget::set_focusable(bool f)
{
	focusable = f;
}

void Widget::set_focus()
{
	if(!parent)
		throw hierarchy_error("no parent");
	if(!visible)
		throw logic_error("!visible");

	signal_request_focus.emit();
}

void Widget::set_state(State mask, State bits)
{
	state = (state&~mask)|bits;
	rebuild();
}

void Widget::rebuild()
{
	if(!style)
		return;

	const Style::PartSeq &parts = style->get_parts();
	list<CachedPart>::iterator j = cached_parts.begin();
	for(Style::PartSeq::const_iterator i=parts.begin(); i!=parts.end(); ++i, ++j)
	{
		if(j==cached_parts.end())
			j = cached_parts.insert(j, CachedPart());
		if(i->get_name().empty())
			i->build(geom, state, *j);
		else
			rebuild_special(*i, *j);
	}
}

void Widget::render(GL::Renderer &renderer) const
{
	if(!style)
		throw logic_error(format("Attempt to render a widget with null style (class=\"%s\", style_name=\"%s\")", get_class(), style_name));

	GL::MatrixStack::Push _pushm(renderer.matrix_stack());
	renderer.matrix_stack() *= GL::Matrix::translation(geom.x, geom.y, 0);
	const Style::PartSeq &parts = style->get_parts();
	list<CachedPart>::const_iterator j = cached_parts.begin();
	for(Style::PartSeq::const_iterator i=parts.begin(); (i!=parts.end() && j!=cached_parts.end()); ++i, ++j)
	{
		if(j->mesh && j->texture)
		{
			renderer.set_texture(j->texture);
			j->mesh->draw(renderer);
		}
		else if(!i->get_name().empty())
			render_special(*i, renderer);
	}
}

void Widget::pointer_enter()
{
	set_state(HOVER);
}

void Widget::pointer_leave()
{
	clear_state(HOVER);
}

void Widget::focus_in()
{
	set_state(FOCUS);
}

void Widget::focus_out()
{
	clear_state(FOCUS);
}


Widget::Loader::Loader(Widget &w):
	DataFile::ObjectLoader<Widget>(w)
{
	add("position", &Loader::position);
	add("size",     &Loader::size);
	add("style",    &Loader::style);
	add("visible",  &Widget::visible);
}

void Widget::Loader::position(int x, int y)
{
	obj.set_position(x, y);
}

void Widget::Loader::size(unsigned w, unsigned h)
{
	obj.set_size(w, h);
}

void Widget::Loader::style(const string &s)
{
	obj.set_style(s);
}

} // namespace GLtk
} // namespace Msp
